@@ -17,6 +17,8 @@ use DBD::Pg;
 use Time::HiRes qw/sleep gettimeofday tv_interval/;
 use Cwd;
 use Data::Dumper;
+use Symbol;
+require Test::More;
 
 use vars qw/$SQL $sth $count $COM %dbh/;
 
@@ -25,7 +27,7 @@ my $DEBUG = $ENV{BUCARDO_DEBUG} || 0;
 $ENV{BUCARDO_CONFIRM} = 0 if exists $ENV{BUCARDO_CONFIRM};
 
 use base 'Exporter';
-our @EXPORT = qw/%tabletype %tabletypemysql %tabletypemariadb %tabletypeoracle %tabletypesqlite
+our @EXPORT = qw/%tabletype %tabletypemysql %tabletypemariadb %tabletypeoracle %tabletypesqlite %tabletypefirebird
                  %sequences %val
                  compare_tables bc_deeply clear_notices wait_for_notice
                  $location $oldherd_msg $newherd_msg $addtable_msg $deltable_msg $nomatch_msg/;
@@ -70,6 +72,33 @@ if (! -e $piddir) {
     mkdir $piddir;
 }
 
+if ($ENV{BUCARDO_LOG_ERROR_CONTEXT}) {
+    no strict 'refs';
+    no warnings qw/prototype redefine/;
+    my ($package) = caller();
+
+    # wrap these routines
+    for my $subname ( qw(ok is like) ) {
+
+        my $glob = qualify_to_ref($subname,$package);
+
+        if (my $sub = *$glob{CODE}) {
+            *$glob = sub {
+                # get result; this is not a general wrapper, since most of
+                # the testing ignores return values here, we aren't worried
+                # about wantarray, etc; we need the return value to decide
+                # if we're going to output a bunch of additional debugging
+                # information.
+                my $res = $sub->( @_ );
+                if (!$res) {
+                    _log_context("@_");
+                }
+                $res;
+            }
+        }
+    }
+}
+
 ## Test databases are labelled as A, B, C, etc.
 my @dbs = qw/A B C D E/;
 
@@ -95,7 +124,7 @@ our %tabletypemysql =
      'bucardo_test1'  => 'SMALLINT',
      'bucardo_test2'  => 'INT',
      'Bucardo_test3'  => 'BIGINT',
-     'bucardo_test4'  => 'VARCHAR(1000)',
+     'bucardo_test4'  => 'VARCHAR(700)',
      'bucardo_test5'  => 'DATE',
      'bucardo_test6'  => 'DATETIME',
      'bucardo_test7'  => 'NUMERIC(5,1)',
@@ -120,6 +149,21 @@ our %tabletypemariadb =
      'bucardo space test' => 'INT',
      );
 
+our %tabletypefirebird =
+    (
+     'bucardo_test1'  => 'SMALLINT',
+     'bucardo_test2'  => 'INT',
+     'Bucardo_test3'  => 'BIGINT',
+     'bucardo_test4'  => 'VARCHAR(700)',
+     'bucardo_test5'  => 'DATE',
+     'bucardo_test6'  => 'DATETIME',
+     'bucardo_test7'  => 'NUMERIC(5,1)',
+     'bucardo_test8'  => 'VARBINARY(1000)',
+     'bucardo_test9'  => 'INTEGER UNSIGNED',
+     'bucardo_test10' => 'TIMESTAMP',
+     'bucardo space test' => 'INT',
+     );
+
 our %tabletypeoracle =
     (
      'bucardo_test1'  => 'SMALLINT',
@@ -194,6 +238,7 @@ else {
     die qq{Could not determine initdb version information from running "$initdb -V"\n};
 }
 
+sub pg_major_version { join '.', $pg_major_version, $pg_minor_version }
 
 ## Each database can also have a custom version
 ## We do this by setting PGBINDIR[A-Z]
@@ -274,7 +319,7 @@ for (1..$xvalmax) {
     $val{SMALLINT}{$_} = $_;
     $val{INT}{$_} = 1234567+$_;
     $val{BIGINT}{$_} = 7777777777 + $_;
-    $val{TEXT}{$_} = $val{'VARCHAR(1000)'}{$_} = "bc'$_";
+    $val{TEXT}{$_} = $val{'VARCHAR(1000)'}{$_} = $val{'VARCHAR(700)'}{$_} = "\\Pbc'$_";
     $val{DATE}{$_} = sprintf '2001-10-%02d', $_;
     $val{TIMESTAMP}{$_} = $val{DATE}{$_} . ' 12:34:56';
     $val{NUMERIC}{$_} = $val{'NUMERIC(5,1)'}{$_} = 0.7 + $_;
@@ -551,13 +596,14 @@ sub start_cluster {
         last if -e $pidfile;
         sleep 0.1;
         if ($loops++ > ($maxwaitseconds * 10)) {
+            Test::More::BAIL_OUT ( 'Failed to connect to database' );
             die "Failed to startup cluster $clustername, command was $COM\n";
         }
         redo;
     }
 
     ## Keep attempting to get a database connection until we get one or timeout
-    $maxwaitseconds = 10;
+    $maxwaitseconds = 20;
 
     my $dbhost = getcwd;
     $dbhost .= "/$dirname/socket";
@@ -1028,6 +1074,7 @@ sub ctl {
     my $dirname = $pgver{A}{dirname};
     $dbhost .= "/$dirname/socket";
     $connopts .= " --dbhost=$dbhost";
+    $connopts .= " --no-bucardorc";
 
     ## Whitespace cleanup
     $args =~ s/^\s+//s;
@@ -1145,16 +1192,19 @@ sub setup_bucardo {
     $self->create_cluster($clustername);
     my $dbh = $self->connect_database($clustername, 'postgres');
     if (database_exists($dbh,'bucardo')) {
-        ## Kick off all other people
+        my $retries = 5;
         my $pidcol = $dbh->{pg_server_version} >= 90200 ? 'pid' : 'procpid';
-        $SQL = qq{SELECT $pidcol FROM pg_stat_activity WHERE datname = 'bucardo' and $pidcol <> pg_backend_pid()};
-        for my $row (@{$dbh->selectall_arrayref($SQL)}) {
-            my $pid = $row->[0];
-            $SQL = 'SELECT pg_terminate_backend(?)';
-            $sth = $dbh->prepare($SQL);
-            $sth->execute($pid);
-        }
-        $dbh->commit();
+        do {
+            ## Kick off all other people
+            $SQL = qq{SELECT $pidcol FROM pg_stat_activity WHERE datname = 'bucardo' and $pidcol <> pg_backend_pid()};
+            for my $row (@{$dbh->selectall_arrayref($SQL)}) {
+                my $pid = $row->[0];
+                $SQL = 'SELECT pg_terminate_backend(?)';
+                $sth = $dbh->prepare($SQL);
+                $sth->execute($pid);
+            }
+            $dbh->commit();
+        } while ($dbh->selectrow_array(qq{SELECT count(*) FROM pg_stat_activity WHERE datname = 'bucardo' and $pidcol <> pg_backend_pid()}))[0] && $retries--;
         debug(qq{Dropping database bucardo from cluster $clustername});
         local $dbh->{AutoCommit} = 1;
         $dbh->do('DROP DATABASE bucardo');
@@ -1194,6 +1244,18 @@ sub setup_bucardo {
 
 } ## end of setup_bucardo
 
+# utility sub called on test error to output pg and bucardo logs to a single
+# output file with context; mainly useful for CI debugging/output
+sub _log_context {
+    return unless $ENV{BUCARDO_LOG_ERROR_CONTEXT};
+
+    warn "Logging context for @_; dir=$ENV{PWD}\n";
+    system("echo '====================' >> log.context");
+    system("date >> log.context");
+    system(sprintf "echo '%s' >> log.context", quotemeta($_[0])) if $_[0];
+    system("tail -n 100 log.bucardo bucardo_test_database_*/pg.log 2>/dev/null >> log.context");
+}
+
 ## Utility functions for object existences:
 sub thing_exists {
     my ($dbh,$name,$table,$column) = @_;
@@ -1883,7 +1945,7 @@ sub check_for_row {
             $dbh->commit();
             if (!$result) {
                 my $line = (caller)[2];
-                Test::More::BAIL_OUT "Stopping on a failed 'check_for_row' test from line $line";
+                Test::More::BAIL_OUT("Stopping on a failed 'check_for_row' test from line $line");
             }
         }
     }
@@ -1972,7 +2034,7 @@ sub is_deeply {
         my $time = time;
         diag("GOT: ".Dumper $_[0]);
         diag("EXPECTED: ".Dumper $_[1]);
-        Test::More::BAIL_OUT "Stopping on a failed 'is_deeply' test from line $line. Time: $time";
+        Test::More::BAIL_OUT("Stopping on a failed 'is_deeply' test from line $line. Time: $time");
     }
 } ## end of is_deeply
 sub like($$;$) {
@@ -1985,7 +2047,7 @@ sub like($$;$) {
         my $time = time;
 #        Test::More::diag("GOT: ".Dumper $_[0]);
 #        Test::More::diag("EXPECTED: ".Dumper $_[1]);
-        Test::More::BAIL_OUT "Stopping on a failed 'like' test from line $line. Time: $time";
+        Test::More::BAIL_OUT("Stopping on a failed 'like' test from line $line. Time: $time");
     }
 } ## end of like
 sub pass(;$) {
@@ -2022,7 +2084,7 @@ sub is($$;$) {
     if ($bail_on_error and ++$total_errors => $bail_on_error) {
         my $line = (caller)[2];
         my $time = time;
-        Test::More::BAIL_OUT "Stopping on a failed 'is' test from line $line. Time: $time";
+        Test::More::BAIL_OUT("Stopping on a failed 'is' test from line $line. Time: $time");
     }
 } ## end of is
 sub isa_ok($$;$) {
@@ -2034,7 +2096,7 @@ sub isa_ok($$;$) {
         return;
     }
     if ($bail_on_error and ++$total_errors => $bail_on_error) {
-        Test::More::BAIL_OUT "Stopping on a failed test";
+        Test::More::BAIL_OUT("Stopping on a failed test");
     }
 } ## end of isa_ok
 sub ok($;$) {
@@ -2045,7 +2107,7 @@ sub ok($;$) {
     if ($bail_on_error and ++$total_errors => $bail_on_error) {
         my $line = (caller)[2];
         my $time = time;
-        Test::More::BAIL_OUT "Stopping on a failed 'ok' test from line $line. Time: $time";
+        Test::More::BAIL_OUT("Stopping on a failed 'ok' test from line $line. Time: $time");
     }
 } ## end of ok
 
